;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; `nokeyb.asm' -- Assembly language source code for `nokeyb.sys'.
;
; This file is part of project `NOKEYB'.
;
; Copyright (c) 2021 Robert Riebisch <rr@bttr-software.de>
;
; Usage of the works is permitted provided that this instrument is retained
; with the works, so that any entity that uses the works is notified of this
; instrument.
;
; DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.
;-----------------------------------------------------------------------------
; Build instructions for NASM version 2.15.05:
;   nasm.exe -o nokeyb.sys nokeyb.asm
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	; limit instruction set
	CPU 8086

	; .SYS file
	ORG 0

deviceDriverHeader:
	DD -1                 ; last driver in file
	;  fedbca9876543210
	DW 1000000000000000b  ; character device driver
	DW strategyRoutine
.iROffset:                    ; patched to `interruptRoutine' by `initialize'
	DW initialize
	DB "NOKEYB$", 0       ; driver name padded to 8 bytes

requestHeaderAddr:
	DD 0

strategyRoutine:
	; just store the address
	mov	[cs:requestHeaderAddr], bx
	mov	[cs:requestHeaderAddr + 2], es
	retf

; device driver status bits
STATUS_ERROR EQU (1 << 15)
STATUS_BUSY  EQU (1 <<  9)
STATUS_DONE  EQU (1 <<  8)

STATUS_SUCCESS EQU !STATUS_ERROR

; device driver error code
ERR_UNKNOWN_COMMAND EQU 3

interruptRoutine:
	; just make all calls fail
	push	bx
	push	ds
	lds	bx, [cs:requestHeaderAddr]
	mov	word [bx + 3], (STATUS_ERROR | STATUS_DONE) + ERR_UNKNOWN_COMMAND
	pop	ds
	pop	bx
	retf

oldInt21hAddr:
	DD 0

myInt21hHandler:
	push	si
	; intercept function `load and execute' only
	cmp	ax, 4b00h
	jne	.notLoadAndExec
	; 'd:\dos\KEYB.COM'
	;  012345678901234
	mov	si, dx
	cmp	word [si], 'd:'
	jne	.notKeyb
	cmp	word [si + 7], 'KE'
	jne	.notKeyb
	cmp	word [si + 9], 'YB'
	jne	.notKeyb
	pop	si
	; intentionally signal error
	stc
	; report as `file not found'
	mov	ax, 2
	; back to caller
	iret
.notLoadAndExec:
.notKeyb:
	pop	si
	; let DOS process all other functions
	jmp	far [cs:oldInt21hAddr]

initialize:
	; save affected registers
	push	ax
	push	bx
	push	dx
	push	ds
	push	es
	pushf

	; print program name and version
	push	cs
	pop	ds
	mov	dx, msgBanner
	mov	ah, 9
	int	21h

	; hook into DOS services interrupt
	mov	ax, (35h << 8) + 21h
	int	21h
	mov	[oldInt21hAddr], bx
	mov	[oldInt21hAddr + 2], es

	mov	dx, myInt21hHandler
	mov	ah, 25h  ; AL still 21h
	int	21h

	; point to dummy routine to keep system state clear
	mov	word [deviceDriverHeader.iROffset], interruptRoutine

	; signal success to DOS
	lds	bx, [requestHeaderAddr]
	mov	word [bx + 3], STATUS_SUCCESS | STATUS_DONE

	; set break address to discard anything above from RAM
	mov	word [bx + 14], initialize
	mov	[bx + 16], cs

	; restore registers
	popf
	pop	es
	pop	ds
	pop	dx
	pop	bx
	pop	ax
	retf

msgBanner:
	DB 'NOKEYB R1', 13, 10
	DB 'Copyright (c) 2021 Robert Riebisch', 13, 10
	DB '$'
